"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const maps_1 = require("../magics/maps");
const utils_1 = require("../utils");
const signaling_1 = require("@phosphor/signaling");
function within_range(position, range) {
    if (range.start.line === range.end.line) {
        return (position.column >= range.start.column &&
            position.column <= range.end.column);
    }
    return ((position.line === range.start.line &&
        position.column >= range.start.column &&
        position.line < range.end.line) ||
        (position.line > range.start.line &&
            position.column <= range.end.column &&
            position.line === range.end.line) ||
        (position.line > range.start.line && position.line < range.end.line));
}
/**
 * A notebook can hold one or more virtual documents; there is always one,
 * "root" document, corresponding to the language of the kernel. All other
 * virtual documents are extracted out of the notebook, based on magics,
 * or other syntax constructs, depending on the kernel language.
 *
 * Virtual documents represent the underlying code in a single language,
 * which has been parsed excluding interactive kernel commands (magics)
 * which could be misunderstood by the specific LSP server.
 *
 * VirtualDocument has no awareness of the notebook or editor it lives in,
 * however it is able to transform its content back to the notebook space,
 * as it keeps editor coordinates for each virtual line.
 *
 * The notebook/editor aware transformations are preferred to be placed in
 * VirtualEditor descendants rather than here.
 */
class VirtualDocument {
    /* Ideas
    signal: on foreign document added
     */
    constructor(language, path, overrides_registry, foreign_code_extractors, standalone, parent) {
        this.path = path;
        this.parent = parent;
        // TODO: make this configurable, depending on the language used
        this.blank_lines_between_cells = 2;
        this.language = language;
        let overrides = language in overrides_registry ? overrides_registry[language] : null;
        this.cell_magics_overrides = new maps_1.CellMagicsMap(overrides ? overrides.cell_magics : []);
        this.line_magics_overrides = new maps_1.LineMagicsMap(overrides ? overrides.line_magics : []);
        this.foreign_extractors_registry = foreign_code_extractors;
        this.foreign_extractors =
            this.language in this.foreign_extractors_registry
                ? this.foreign_extractors_registry[this.language]
                : [];
        this.virtual_lines = new Map();
        this.source_lines = new Map();
        this.foreign_documents = new Map();
        this.overrides_registry = overrides_registry;
        this.lines = [];
        this.standalone = standalone;
        this.instance_id = VirtualDocument.instances_count;
        VirtualDocument.instances_count += 1;
        this.unused_standalone_documents = new utils_1.DefaultMap(() => new Array());
        this._remaining_lifetime = 10;
        this.foreign_document_closed = new signaling_1.Signal(this);
        this.foreign_document_opened = new signaling_1.Signal(this);
        this._shift = {
            start: { line: 0, column: 0 },
            end: { line: 0, column: 0 }
        };
    }
    /**
     * When this counter goes down to 0, the document will be destroyed and the associated connection will be closed;
     * This is meant to reduce the number of open connections when a a foreign code snippet was removed from the document.
     *
     * Note: top level virtual documents are currently immortal (unless killed by other means); it might be worth
     * implementing culling of unused documents, but if and only if JupyterLab will also implement culling of
     * idle kernels - otherwise the user experience could be a bit inconsistent, and we would need to invent our own rules.
     */
    get remaining_lifetime() {
        if (!this.parent) {
            return Infinity;
        }
        return this._remaining_lifetime;
    }
    set remaining_lifetime(value) {
        if (this.parent) {
            this._remaining_lifetime = value;
        }
    }
    clear() {
        // TODO - deep clear (assure that there is no memory leak)
        this.unused_standalone_documents.clear();
        for (let document of this.foreign_documents.values()) {
            document.clear();
            if (document.standalone) {
                // once the standalone document was cleared, we may want to remove it and close connection;
                // but wait, this is a waste of resources (opening a connection takes 1-3 seconds) and,
                // since this is cleaned anyway, we could use it for another standalone document of the same language.
                let set = this.unused_standalone_documents.get(document.language);
                set.push(document);
            }
        }
        this.unused_documents = new Set(this.foreign_documents.values());
        this.virtual_lines.clear();
        this.source_lines.clear();
        this.last_virtual_line = 0;
        this.last_source_line = 0;
        this.lines = [];
    }
    forward_closed_signal(host, context) {
        this.foreign_document_closed.emit(context);
    }
    forward_opened_signal(host, context) {
        this.foreign_document_opened.emit(context);
    }
    // TODO: what could be refactored into "ForeignDocumentsManager" has started to emerge;
    //   we should consider refactoring later on.
    open_foreign(language, standalone) {
        let document = new VirtualDocument(language, this.path, this.overrides_registry, this.foreign_extractors_registry, standalone, this);
        this.foreign_document_opened.emit({
            foreign_document: document,
            parent_host: this
        });
        // pass through any future signals
        document.foreign_document_closed.connect(this.forward_closed_signal);
        document.foreign_document_opened.connect(this.forward_opened_signal);
        this.foreign_documents.set(document.virtual_id, document);
        return document;
    }
    document_at_source_position(position) {
        let source_line = this.source_lines.get(position.line);
        let source_position_ce = {
            line: source_line.editor_line,
            column: position.ch
        };
        for (let [range, document] of source_line.foreign_documents_map) {
            // console.log('Range', range)
            if (within_range(source_position_ce, range)) {
                let source_position_cm = {
                    line: source_position_ce.line - range.start.line,
                    ch: source_position_ce.column - range.start.column
                };
                return document.document_at_source_position(source_position_cm);
            }
        }
        return this;
    }
    virtual_position_at_document(source_position) {
        let source_line = this.source_lines.get(source_position.line);
        let virtual_line = source_line.virtual_line;
        let source_position_ce = {
            line: source_line.editor_line,
            column: source_position.ch
        };
        for (let [range, document] of source_line.foreign_documents_map) {
            // console.log('Range', range)
            if (within_range(source_position_ce, range)) {
                let source_position_cm = {
                    line: source_position_ce.line - range.start.line,
                    ch: source_position_ce.column - range.start.column
                };
                return document.virtual_position_at_document(source_position_cm);
            }
        }
        return {
            ch: source_position.ch,
            line: virtual_line
        };
    }
    append_code_block(cell_code, cm_editor) {
        let source_cell_lines = cell_code.split('\n');
        let lines;
        let should_inspect;
        //  TODO: create additional LSP servers for foreign languages introduced in cell magics
        let foreign_document_map = new Map();
        // first, check if there is any foreign code:
        // let is_line_foreign = new Array<boolean>();
        for (let extractor of this.foreign_extractors) {
            let result = extractor.extract_foreign_code(cell_code);
            if (result.foreign_code === null) {
                continue;
            }
            let foreign_document;
            // if not standalone, try to append to existing document
            let foreign_exists = this.foreign_documents.has(extractor.language);
            if (!extractor.standalone && foreign_exists) {
                foreign_document = this.foreign_documents.get(extractor.language);
                this.unused_documents.delete(foreign_document);
            }
            else {
                // if standalone, try to re-use existing connection to the server
                let unused_standalone = this.unused_standalone_documents.get(extractor.language);
                if (extractor.standalone && unused_standalone.length > 0) {
                    foreign_document = unused_standalone.pop();
                }
                else {
                    // if (previous document does not exists) or (extractor produces standalone documents
                    // and no old standalone document could be reused): create a new document
                    foreign_document = this.open_foreign(extractor.language, extractor.standalone);
                }
            }
            foreign_document.append_code_block(result.foreign_code, cm_editor);
            cell_code = result.host_code;
            foreign_document_map.set(result.range, foreign_document);
            // TODO to constructor!, TODO remove map?
            foreign_document._shift = result.range;
            // not breaking - many extractors are allowed to process the code, one after each other
            // (think JS and CSS in HTML, or %R inside of %%timeit).
        }
        // cell magics are replaced if requested and matched
        let cell_override = this.cell_magics_overrides.override_for(cell_code);
        if (cell_override !== null) {
            lines = cell_override.split('\n');
            should_inspect = lines.map(l => false);
        }
        else {
            // otherwise, we replace line magics - if any
            let result = this.line_magics_overrides.replace_all(cell_code.split('\n'));
            lines = result.lines;
            should_inspect = result.should_inspect;
        }
        for (let i = 0; i < lines.length; i++) {
            this.virtual_lines.set(this.last_virtual_line + i, {
                inspect: should_inspect[i],
                editor: cm_editor,
                // TODO this is incorrect, wont work if something was extracted
                source_line: this.last_source_line + i
            });
        }
        for (let i = 0; i < source_cell_lines.length; i++) {
            this.source_lines.set(this.last_source_line + i, {
                editor_line: i,
                editor: cm_editor,
                foreign_documents_map: foreign_document_map,
                // TODO this is incorrect, wont work if something was extracted
                virtual_line: this.last_virtual_line + i
            });
        }
        // one empty line is necessary to separate code blocks, next 'n' lines are to silence linters;
        // the final cell does not get the additional lines (thanks to the use of join, see below)
        this.lines.push(lines.join('\n') + '\n');
        this.last_virtual_line += lines.length + this.blank_lines_between_cells;
        this.last_source_line += source_cell_lines.length;
    }
    get value() {
        let lines_padding = '\n'.repeat(this.blank_lines_between_cells);
        // TODO: require the editor to do this
        // once all the foreign documents were refreshed, the unused documents (and their connections)
        // should be terminated if their lifetime expired
        this.close_expired_documents();
        return this.lines.join(lines_padding);
    }
    close_expired_documents() {
        for (let document of this.unused_documents.values()) {
            document.remaining_lifetime -= 1;
            if (document.remaining_lifetime <= 0) {
                this.close_foreign(document);
            }
        }
    }
    close_foreign(document) {
        console.log('Closing document', document);
        this.foreign_document_closed.emit({
            foreign_document: document,
            parent_host: this
        });
        // remove it from foreign documents list
        this.foreign_documents.delete(document.virtual_id);
        // and delete the documents within it
        document.close_all_foreign_documents();
        document.foreign_document_closed.disconnect(this.forward_closed_signal);
        document.foreign_document_opened.disconnect(this.forward_opened_signal);
    }
    close_all_foreign_documents() {
        for (let document of this.foreign_documents.values()) {
            this.close_foreign(document);
        }
    }
    get virtual_id() {
        // for easier debugging, the language information is included in the ID:
        return this.standalone
            ? this.instance_id + '(' + this.language + ')'
            : this.language;
    }
    get id_path() {
        if (!this.parent) {
            return this.virtual_id;
        }
        return this.parent.id_path + '-' + this.virtual_id;
    }
    get uri() {
        return this.path + '.' + this.id_path;
    }
    transform_source_to_editor(pos) {
        // console.log(this.source_lines)
        return Object.assign({}, pos, { line: this.source_lines.get(pos.line).editor_line, is_editor: true });
    }
    transform_virtual_to_editor(virtual_position) {
        return this.unshift(this.transform_source_to_editor(this.transform_virtual_to_source(virtual_position)));
    }
    transform_virtual_to_source(position) {
        return Object.assign({}, position, { line: this.virtual_lines.get(position.line).source_line });
    }
    /*
    shift(pos: CodeMirror.Position): CodeMirror.Position {
      let transformed = {
        line: pos.line - this._shift.start.line,
        ch: pos.ch - (pos.line === 0 ? this._shift.start.column : 0)
      };
      if (this.parent) {
        return this.parent.shift(transformed);
      }
      return transformed;
    }
     */
    unshift(pos) {
        // TODO: shift and unshift should be done in the offset space?
        let transformed = Object.assign({}, pos, { line: pos.line + this._shift.start.line, ch: pos.ch + (pos.line === 0 ? this._shift.start.column : 0) });
        if (this.parent) {
            return this.parent.unshift(transformed);
        }
        return transformed;
    }
    get root() {
        if (typeof this.parent === 'undefined') {
            return this;
        }
        return this.parent.root;
    }
    get_editor_at_virtual_line(pos) {
        return this.virtual_lines.get(pos.line).editor;
    }
    get_editor_at_source_line(pos) {
        return this.source_lines.get(pos.line).editor;
    }
}
VirtualDocument.instances_count = 0;
exports.VirtualDocument = VirtualDocument;
